@@ -2,7 +2,7 @@
 /**
  * This file contains the Net_Sieve class.
  *
- * PHP version 4
+ * PHP version 5
  *
  * +-----------------------------------------------------------------------+
  * | All rights reserved.                                                  |
@@ -39,33 +39,29 @@
  * @copyright 2002-2003 Richard Heyes
  * @copyright 2006-2008 Anish Mistry
  * @license   http://www.opensource.org/licenses/bsd-license.php BSD
- * @version   SVN: $Id$
  * @link      http://pear.php.net/package/Net_Sieve
  */
 
 require_once 'PEAR.php';
 require_once 'Net/Socket.php';
 
-/**
- * TODO
- *
- * o supportsAuthMech()
- */
-
 /**
  * Disconnected state
+ *
  * @const NET_SIEVE_STATE_DISCONNECTED
  */
 define('NET_SIEVE_STATE_DISCONNECTED', 1, true);
 
 /**
  * Authorisation state
+ *
  * @const NET_SIEVE_STATE_AUTHORISATION
  */
 define('NET_SIEVE_STATE_AUTHORISATION', 2, true);
 
 /**
  * Transaction state
+ *
  * @const NET_SIEVE_STATE_TRANSACTION
  */
 define('NET_SIEVE_STATE_TRANSACTION', 3, true);
@@ -80,10 +76,11 @@ define('NET_SIEVE_STATE_TRANSACTION', 3, true);
  * @author    Damian Fernandez Sosa <damlists@cnba.uba.ar>
  * @author    Anish Mistry <amistry@am-productions.biz>
  * @author    Jan Schneider <jan@horde.org>
+ * @author    Neil Munday <neil@mundayweb.com>
  * @copyright 2002-2003 Richard Heyes
  * @copyright 2006-2008 Anish Mistry
  * @license   http://www.opensource.org/licenses/bsd-license.php BSD
- * @version   Release: 1.3.3
+ * @version   Release: 1.4.3
  * @link      http://pear.php.net/package/Net_Sieve
  * @link      http://tools.ietf.org/html/rfc5228 RFC 5228 (Sieve: An Email
  *            Filtering Language)
@@ -99,8 +96,14 @@ class Net_Sieve
      *
      * @var array
      */
-    var $supportedAuthMethods = array('DIGEST-MD5', 'CRAM-MD5', 'EXTERNAL',
-                                      'PLAIN' , 'LOGIN');
+    var $supportedAuthMethods = array(
+        'DIGEST-MD5',
+        'CRAM-MD5',
+        'EXTERNAL',
+        'PLAIN' ,
+        'LOGIN',
+        'GSSAPI'
+    );
 
     /**
      * SASL authentication methods that require Auth_SASL.
@@ -190,6 +193,20 @@ class Net_Sieve
      */
     var $_maxReferralCount = 15;
 
+    /**
+     * Kerberos service principal to use for GSSAPI authentication.
+     *
+     * @var string
+     */
+    var $_gssapiPrincipal = null;
+
+    /**
+     * Kerberos service cname to use for GSSAPI authentication.
+     *
+     * @var string
+     */
+    var $_gssapiCN = null;
+
     /**
      * Constructor.
      *
@@ -212,12 +229,16 @@ class Net_Sieve
      * @param array   $options    Additional options for
      *                            stream_context_create().
      * @param mixed   $handler    A callback handler for the debug output.
+     * @param string  $principal  Kerberos service principal to use
+     *                            with GSSAPI authentication.
+     * @param string  $cname      Kerberos service cname to use
+     *                            with GSSAPI authentication.
      */
-    function Net_Sieve($user = null, $pass  = null, $host = 'localhost',
-                       $port = 2000, $logintype = '', $euser = '',
-                       $debug = false, $bypassAuth = false, $useTLS = true,
-                       $options = null, $handler = null)
-    {
+    function __construct($user = null, $pass  = null, $host = 'localhost',
+        $port = 2000, $logintype = '', $euser = '',
+        $debug = false, $bypassAuth = false, $useTLS = true,
+        $options = null, $handler = null, $principal = null, $cname = null
+    ) {
         $this->_pear = new PEAR();
         $this->_state             = NET_SIEVE_STATE_DISCONNECTED;
         $this->_data['user']      = $user;
@@ -229,7 +250,10 @@ class Net_Sieve
         $this->_sock              = new Net_Socket();
         $this->_bypassAuth        = $bypassAuth;
         $this->_useTLS            = $useTLS;
-        $this->_options           = $options;
+        $this->_options           = (array) $options;
+        $this->_gssapiPrincipal   = $principal;
+        $this->_gssapiCN          = $cname;
+
         $this->setDebug($debug, $handler);
 
         /* Try to include the Auth_SASL package.  If the package is not
@@ -272,6 +296,32 @@ class Net_Sieve
         $this->_debug_handler = $handler;
     }
 
+    /**
+     * Sets the Kerberos service principal for use with GSSAPI
+     * authentication.
+     *
+     * @param string $principal The Kerberos service principal
+     *
+     * @return void
+     */
+    function setServicePrincipal($principal)
+    {
+        $this->_gssapiPrincipal = $principal;
+    }
+
+    /**
+     * Sets the Kerberos service CName for use with GSSAPI
+     * authentication.
+     *
+     * @param string $cname The Kerberos service principal
+     *
+     * @return void
+     */
+    function setServiceCN($cname)
+    {
+        $this->_gssapiCN = $cname;
+    }
+
     /**
      * Connects to the server and logs in.
      *
@@ -308,6 +358,7 @@ class Net_Sieve
         $this->_data['host'] = $host;
         $this->_data['port'] = $port;
         $this->_useTLS       = $useTLS;
+
         if (is_array($options)) {
             $this->_options = array_merge($this->_options, $options);
         }
@@ -323,21 +374,29 @@ class Net_Sieve
 
         if ($this->_bypassAuth) {
             $this->_state = NET_SIEVE_STATE_TRANSACTION;
+
+            // Reset capabilities
+            $this->_parseCapability('');
         } else {
             $this->_state = NET_SIEVE_STATE_AUTHORISATION;
+
             $res = $this->_doCmd();
             if (is_a($res, 'PEAR_Error')) {
                 return $res;
             }
+
+            // Reset capabilities (use unattended capabilities)
+            $this->_parseCapability($res);
         }
 
-        // Explicitly ask for the capabilities in case the connection is
-        // picked up from an existing connection.
-        $res = $this->_cmdCapability();
-        if (is_a($res, 'PEAR_Error')) {
-            return $this->_pear->raiseError(
-                'Failed to connect, server said: ' . $res->getMessage(), 2
-            );
+        // Explicitly ask for the capabilities if needed
+        if (empty($this->_capability['implementation'])) {
+            $res = $this->_cmdCapability();
+            if (is_a($res, 'PEAR_Error')) {
+                return $this->_pear->raiseError(
+                    'Failed to connect, server said: ' . $res->getMessage(), 2
+                );
+            }
         }
 
         // Check if we can enable TLS via STARTTLS.
@@ -395,6 +454,7 @@ class Net_Sieve
                 return $res;
             }
         }
+
         $this->_state = NET_SIEVE_STATE_TRANSACTION;
 
         return true;
@@ -403,16 +463,21 @@ class Net_Sieve
     /**
      * Returns an indexed array of scripts currently on the server.
      *
-     * @return array  Indexed array of scriptnames.
+     * @param string $active Will be set to the name of the active script
+     *
+     * @return array  Indexed array of scriptnames, PEAR_Error on failure
      */
-    function listScripts()
+    function listScripts(&$active = null)
     {
         if (is_array($scripts = $this->_cmdListScripts())) {
-            $this->_active = $scripts[1];
+            if (isset($scripts[1])) {
+                $active = $scripts[1];
+            }
+
             return $scripts[0];
-        } else {
-            return $scripts;
         }
+
+        return $scripts;
     }
 
     /**
@@ -422,11 +487,7 @@ class Net_Sieve
      */
     function getActive()
     {
-        if (!empty($this->_active)) {
-            return $this->_active;
-        }
         if (is_array($scripts = $this->_cmdListScripts())) {
-            $this->_active = $scripts[1];
             return $scripts[1];
         }
     }
@@ -470,9 +531,11 @@ class Net_Sieve
         if (is_a($res, 'PEAR_Error')) {
             return $res;
         }
+
         if ($makeactive) {
             return $this->_cmdSetActive($scriptname);
         }
+
         return true;
     }
 
@@ -505,9 +568,10 @@ class Net_Sieve
         }
 
         $res = $this->_doCmd(sprintf('HAVESPACE %s %d', $this->_escape($scriptname), $size));
-        if (is_a($res, 'PEAR_Error')) {        
+        if (is_a($res, 'PEAR_Error')) {
             return $res;
         }
+
         return true;
     }
 
@@ -521,6 +585,7 @@ class Net_Sieve
         if (NET_SIEVE_STATE_DISCONNECTED == $this->_state) {
             return $this->_pear->raiseError('Not currently connected', 7);
         }
+
         return $this->_capability['extensions'];
     }
 
@@ -560,6 +625,7 @@ class Net_Sieve
         if (NET_SIEVE_STATE_DISCONNECTED == $this->_state) {
             return $this->_pear->raiseError('Not currently connected', 7);
         }
+
         return $this->_capability['sasl'];
     }
 
@@ -578,6 +644,7 @@ class Net_Sieve
         }
 
         $method = trim($this->_toUpper($method));
+
         if (is_array($this->_capability['sasl'])) {
             foreach ($this->_capability['sasl'] as $sasl) {
                 if ($sasl == $method) {
@@ -606,6 +673,7 @@ class Net_Sieve
         if (is_a($method, 'PEAR_Error')) {
             return $method;
         }
+
         switch ($method) {
         case 'DIGEST-MD5':
             return $this->_authDigestMD5($uid, $pwd, $euser);
@@ -621,6 +689,9 @@ class Net_Sieve
         case 'EXTERNAL':
             $result = $this->_authEXTERNAL($uid, $pwd, $euser);
             break;
+        case 'GSSAPI':
+            $result = $this->_authGSSAPI($pwd);
+            break;
         default :
             $result = $this->_pear->raiseError(
                 $method . ' is not a supported authentication method'
@@ -633,7 +704,6 @@ class Net_Sieve
             return $res;
         }
 
-        // Query the server capabilities again now that we are authenticated.
         if ($this->_pear->isError($res = $this->_cmdCapability())) {
             return $this->_pear->raiseError(
                 'Failed to connect, server said: ' . $res->getMessage(), 2
@@ -662,12 +732,72 @@ class Net_Sieve
         );
     }
 
+    /**
+     * Authenticates the user using the GSSAPI method.
+     *
+     * @note the PHP krb5 extension is required and the service principal and cname
+     *       must have been set.
+     * @see  setServicePrincipal()
+     *
+     * @return void
+     */
+    function _authGSSAPI()
+    {
+        if (!extension_loaded('krb5')) {
+            return $this->_pear->raiseError('The krb5 extension is required for GSSAPI authentication', 2);
+        }
+
+        if (!$this->_gssapiPrincipal) {
+            return $this->_pear->raiseError('No Kerberos service principal set', 2);
+        }
+
+        if (!$this->_gssapiCN) {
+            return $this->_pear->raiseError('No Kerberos service CName set', 2);
+        }
+
+        putenv('KRB5CCNAME=' . $this->_gssapiCN);
+
+        try {
+            $ccache = new KRB5CCache();
+            $ccache->open($this->_gssapiCN);
+
+            $gssapicontext = new GSSAPIContext();
+            $gssapicontext->acquireCredentials($ccache);
+
+            $token   = '';
+            $success = $gssapicontext->initSecContext($this->_gssapiPrincipal, null, null, null, $token);
+            $token   = base64_encode($token);
+        }
+        catch (Exception $e) {
+            return $this->_pear->raiseError('GSSAPI authentication failed: ' . $e->getMessage());
+        }
+
+        $this->_sendCmd("AUTHENTICATE \"GSSAPI\" {" . strlen($token) . "+}");
+
+        $response = $this->_doCmd($token, true);
+
+        try {
+            $challenge = base64_decode(substr($response, 1, -1));
+            $gssapicontext->unwrap($challenge, $challenge);
+            $gssapicontext->wrap($challenge, $challenge, true);
+        }
+        catch (Exception $e) {
+            return $this->_pear->raiseError('GSSAPI authentication failed: ' . $e->getMessage());
+        }
+
+        $response = base64_encode($challenge);
+
+        $this->_sendCmd("{" . strlen($response) . "+}");
+
+        return $this->_sendCmd($response);
+    }
+
     /**
      * Authenticates the user using the LOGIN method.
      *
      * @param string $user  The userid to authenticate as.
      * @param string $pass  The password to authenticate with.
-     * @param string $euser The effective uid to authenticate as.
+     * @param string $euser The effective uid to authenticate as. Not used.
      *
      * @return void
      */
@@ -677,10 +807,12 @@ class Net_Sieve
         if (is_a($result, 'PEAR_Error')) {
             return $result;
         }
+
         $result = $this->_doCmd('"' . base64_encode($user) . '"', true);
         if (is_a($result, 'PEAR_Error')) {
             return $result;
         }
+
         return $this->_doCmd('"' . base64_encode($pass) . '"', true);
     }
 
@@ -689,7 +821,7 @@ class Net_Sieve
      *
      * @param string $user  The userid to authenticate as.
      * @param string $pass  The password to authenticate with.
-     * @param string $euser The effective uid to authenticate as.
+     * @param string $euser The effective uid to authenticate as. Not used.
      *
      * @return void
      */
@@ -700,9 +832,11 @@ class Net_Sieve
             return $challenge;
         }
 
+        $auth_sasl = new Auth_SASL;
+        $cram      = $auth_sasl->factory('crammd5');
         $challenge = base64_decode(trim($challenge));
-        $cram = Auth_SASL::factory('crammd5');
-        $response = $cram->getResponse($user, $pass, $challenge);
+        $response  = $cram->getResponse($user, $pass, $challenge);
+
         if (is_a($response, 'PEAR_Error')) {
             return $response;
         }
@@ -726,8 +860,10 @@ class Net_Sieve
             return $challenge;
         }
 
+        $auth_sasl = new Auth_SASL;
+        $digest    = $auth_sasl->factory('digestmd5');
         $challenge = base64_decode(trim($challenge));
-        $digest = Auth_SASL::factory('digestmd5');
+
         // @todo Really 'localhost'?
         $response = $digest->getResponse($user, $pass, $challenge, 'localhost', 'sieve', $euser);
         if (is_a($response, 'PEAR_Error')) {
@@ -738,10 +874,12 @@ class Net_Sieve
         if (is_a($result, 'PEAR_Error')) {
             return $result;
         }
+
         $result = $this->_doCmd('', true);
         if (is_a($result, 'PEAR_Error')) {
             return $result;
         }
+
         if ($this->_toUpper(substr($result, 0, 2)) == 'OK') {
             return;
         }
@@ -765,7 +903,7 @@ class Net_Sieve
      *
      * @return void
      *
-     * @since  1.1.7
+     * @since 1.1.7
      */
     function _authEXTERNAL($user, $pass, $euser)
     {
@@ -773,6 +911,7 @@ class Net_Sieve
             'AUTHENTICATE "EXTERNAL" "%s"',
             base64_encode(strlen($euser) ? $euser : $user)
         );
+
         return $this->_sendCmd($cmd);
     }
 
@@ -793,6 +932,7 @@ class Net_Sieve
         if (is_a($res, 'PEAR_Error')) {
             return $res;
         }
+
         return true;
     }
 
@@ -836,7 +976,6 @@ class Net_Sieve
             return $res;
         }
 
-        $this->_activeScript = $scriptname;
         return true;
     }
 
@@ -889,10 +1028,13 @@ class Net_Sieve
         }
 
         $stringLength = $this->_getLineLength($scriptdata);
-        $command      = sprintf("PUTSCRIPT %s {%d+}\r\n%s",
-                                $this->_escape($scriptname),
-                                $stringLength,
-                                $scriptdata);
+        $command      = sprintf(
+            "PUTSCRIPT %s {%d+}\r\n%s",
+            $this->_escape($scriptname),
+            $stringLength,
+            $scriptdata
+        );
+
         $res = $this->_doCmd($command);
         if (is_a($res, 'PEAR_Error')) {
             return $res;
@@ -972,11 +1114,15 @@ class Net_Sieve
                 break;
 
             case 'SASL':
-                $this->_capability['sasl'] = preg_split('/\s+/', $matches[3]);
+                if (!empty($matches[3])) {
+                    $this->_capability['sasl'] = preg_split('/\s+/', $matches[3]);
+                }
                 break;
 
             case 'SIEVE':
-                $this->_capability['extensions'] = preg_split('/\s+/', $matches[3]);
+                if (!empty($matches[3])) {
+                    $this->_capability['extensions'] = preg_split('/\s+/', $matches[3]);
+                }
                 break;
 
             case 'STARTTLS':
@@ -1047,9 +1193,9 @@ class Net_Sieve
     /**
      * Receives a number of bytes from the server.
      *
-     * @param integer $length  Number of bytes to read.
+     * @param integer $length Number of bytes to read.
      *
-     * @return string  The server response.
+     * @return string The server response.
      */
     function _recvBytes($length)
     {
@@ -1066,11 +1212,11 @@ class Net_Sieve
     /**
      * Send a command and retrieves a response from the server.
      *
-     * @param string $cmd   The command to send.
+     * @param string  $cmd  The command to send.
      * @param boolean $auth Whether this is an authentication command.
      *
-     * @return string|PEAR_Error  Reponse string if an OK response, PEAR_Error
-     *                            if a NO response.
+     * @return string|PEAR_Error Reponse string if an OK response, PEAR_Error
+     *                           if a NO response.
      */
     function _doCmd($cmd = '', $auth = false)
     {
@@ -1188,10 +1334,11 @@ class Net_Sieve
             if (in_array($userMethod, $this->_capability['sasl'])) {
                 return $userMethod;
             }
+
+            $msg = 'No supported authentication method found. The server supports these methods: %s, but we want to use: %s';
             return $this->_pear->raiseError(
-                sprintf('No supported authentication method found. The server supports these methods: %s, but we want to use: %s',
-                        implode(', ', $this->_capability['sasl']),
-                        $userMethod));
+                sprintf($msg, implode(', ', $this->_capability['sasl']), $userMethod)
+            );
         }
 
         foreach ($this->supportedAuthMethods as $method) {
@@ -1200,10 +1347,10 @@ class Net_Sieve
             }
         }
 
+        $msg = 'No supported authentication method found. The server supports these methods: %s, but we only support: %s';
         return $this->_pear->raiseError(
-            sprintf('No supported authentication method found. The server supports these methods: %s, but we only support: %s',
-                    implode(', ', $this->_capability['sasl']),
-                    implode(', ', $this->supportedAuthMethods)));
+            sprintf($msg, implode(', ', $this->_capability['sasl']), implode(', ', $this->supportedAuthMethods))
+        );
     }
 
     /**
@@ -1218,7 +1365,17 @@ class Net_Sieve
             return $res;
         }
 
-        if (!stream_socket_enable_crypto($this->_sock->fp, true, STREAM_CRYPTO_METHOD_TLS_CLIENT)) {
+        if (isset($this->_options['ssl']['crypto_method'])) {
+            $crypto_method = $this->_options['ssl']['crypto_method'];
+        } else {
+            // There is no flag to enable all TLS methods. Net_SMTP
+            // handles enabling TLS similarly.
+            $crypto_method = STREAM_CRYPTO_METHOD_TLS_CLIENT
+                | @STREAM_CRYPTO_METHOD_TLSv1_1_CLIENT
+                | @STREAM_CRYPTO_METHOD_TLSv1_2_CLIENT;
+        }
+
+        if (!stream_socket_enable_crypto($this->_sock->fp, true, $crypto_method)) {
             return $this->_pear->raiseError('Failed to establish TLS connection', 2);
         }
 
@@ -1229,18 +1386,23 @@ class Net_Sieve
         // Unfortunately old Cyrus versions are broken and don't send a
         // CAPABILITY response, thus we would wait here forever. Parse the
         // Cyrus version and work around this broken behavior.
-        if (!preg_match('/^CYRUS TIMSIEVED V([0-9.]+)/', $this->_capability['implementation'], $matches) ||
-            version_compare($matches[1], '2.3.10', '>=')) {
-            $this->_doCmd();
+        if (!preg_match('/^CYRUS TIMSIEVED V([0-9.]+)/', $this->_capability['implementation'], $matches)
+            || version_compare($matches[1], '2.3.10', '>=')
+        ) {
+            $res = $this->_doCmd();
         }
 
-        // Query the server capabilities again now that we are under
-        // encryption.
-        $res = $this->_cmdCapability();
-        if (is_a($res, 'PEAR_Error')) {
-            return $this->_pear->raiseError(
-                'Failed to connect, server said: ' . $res->getMessage(), 2
-            );
+        // Reset capabilities (use unattended capabilities)
+        $this->_parseCapability(is_string($res) ? $res : '');
+
+        // Query the server capabilities again now that we are under encryption.
+        if (empty($this->_capability['implementation'])) {
+            $res = $this->_cmdCapability();
+            if (is_a($res, 'PEAR_Error')) {
+                return $this->_pear->raiseError(
+                    'Failed to connect, server said: ' . $res->getMessage(), 2
+                );
+            }
         }
 
         return true;
@@ -1256,7 +1418,7 @@ class Net_Sieve
     function _getLineLength($string)
     {
         if (extension_loaded('mbstring')) {
-            return mb_strlen($string, 'latin1');
+            return mb_strlen($string, '8bit');
         } else {
             return strlen($string);
         }
@@ -1281,9 +1443,9 @@ class Net_Sieve
     /**
      * Converts strings into RFC's quoted-string or literal-c2s form.
      *
-     * @param string $string  The string to convert.
+     * @param string $string The string to convert.
      *
-     * @return string  Result string.
+     * @return string Result string.
      */
     function _escape($string)
     {
